# =====================================
# generator=datazen
# version=3.2.3
# hash=8042bfab9549cff58935b1dd444a1e8f
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "4.5.0"
